use png;
use rand::Rng;
use std::collections::HashSet;
use std::fmt;
use std::fs::File;
use std::io::prelude::*;
use std::io::{BufRead, BufReader, BufWriter};
use std::path::Path;
use std::result::Result;
use structopt::StructOpt;

/// Generates Game Of Life wallpapers.
#[derive(StructOpt, Debug)]
#[structopt(name = "gol")]
struct Opt {
    /// Width of the wallpaper
    #[structopt(name = "width")]
    width: u32,

    /// Height of the wallpaper
    #[structopt(name = "height")]
    height: u32,
    // /// bla
    // #[structopt(name = "pixelsize")]
    // pixel_size: u32,
}

#[derive(Debug)]
pub enum FormatError {
    ParserError,
}
impl fmt::Display for FormatError {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        match *self {
            _ => f.write_str("Error while parsing."),
        }
    }
}

impl std::error::Error for FormatError {
    fn description(&self) -> &str {
        match *self {
            _ => "Error while parsing.",
        }
    }
}

struct Field {
    width: u32,
    height: u32,
    tmp_path: std::string::String,
    data: HashSet<(u32, u32)>,
}

impl Field {
    fn with_size(width: u32, height: u32) -> Self {
        //fn with_size(width: u32, height: u32, pixel_size: u32) -> Self {
        Field {
            width,
            height,
            tmp_path: "/tmp/wallpaper.gol".to_string(),
            data: HashSet::new(),
        }
    }

    fn write_file(&self) {
        let mut output = "".to_string();
        for coords in &self.data {
            output.push_str(&format!("{}/{}\n", coords.0, coords.1));
        }
        let mut file = File::create(&self.tmp_path).unwrap();
        write!(file, "{}", output).unwrap();
    }

    fn read_file(&mut self) -> Result<(), FormatError> {
        let file = File::open(&self.tmp_path);
        if file.is_err() {
            return Err(FormatError::ParserError);
        }
        let file = BufReader::new(file.unwrap());
        self.data = HashSet::new();
        for line in file.lines() {
            if line.is_err() {
                return Err(FormatError::ParserError);
            }
            let line = line.unwrap();

            let parts: Vec<_> = line.split('/').collect();
            let x = parts[0].parse();
            let y = parts[1].parse();
            if x.is_err() || y.is_err() {
                return Err(FormatError::ParserError);
            }
            let x: u32 = x.unwrap();
            let y: u32 = y.unwrap();
            self.data.insert((x, y));
        }
        Ok(())
    }

    fn write_image(&self, path: &str) {
        let path = Path::new(&path);
        let file = File::create(path).unwrap();
        let ref mut w = BufWriter::new(file);

        let mut encoder = png::Encoder::new(w, self.width, self.height);
        encoder.set_color(png::ColorType::RGB);
        encoder.set_depth(png::BitDepth::Eight);
        let mut writer = encoder.write_header().unwrap();

        let mut pixels = Vec::<u8>::new();
        for y in 0..self.height {
            for x in 0..self.width {
                if self.data.contains(&(x, y)) {
                    pixels.push(255);
                    pixels.push(0);
                    pixels.push(0);
                } else {
                    pixels.push(0);
                    pixels.push(0);
                    pixels.push(0);
                }
            }
        }
        // println!("{:?}", pixels);

        //let data = [255, 0, 0, 0, 0, 0, 255, 0, 0, 0, 0, 0, 255, 0, 0, 0, 0, 0, 255, 0, 0, 0, 0, 0, 255, 0, 0, 0, 0, 0];
        writer.write_image_data(pixels.as_slice()).unwrap(); // Save
    }

    fn fill_random(&mut self) {
        let mut rng = rand::thread_rng();
        for x in 0..self.width {
            for y in 0..self.height {
                let r = rng.gen_range(0, 100);
                if r < 15 {
                    self.data.insert((x, y));
                }
            }
        }
    }

    fn count_neighbours(&self, x: u32, y: u32) -> u8 {
        let mut neighbours = 0;

        if x > 0 && y > 0 {
            neighbours += self.is_alive(x - 1, y - 1);
        }

        if x > 0 {
            neighbours += self.is_alive(x - 1, y);
            neighbours += self.is_alive(x - 1, y + 1);
        }

        if y > 0 {
            neighbours += self.is_alive(x, y - 1);
            neighbours += self.is_alive(x + 1, y - 1);
        }

        neighbours += self.is_alive(x + 1, y);
        neighbours += self.is_alive(x, y + 1);
        neighbours += self.is_alive(x + 1, y + 1);

        //if x == 2 && (y == 0 || y == 2) {
        //println!("{}/{}: {}", x, y, neighbours);
        //}
        neighbours
    }

    fn is_alive(&self, x: u32, y: u32) -> u8 {
        if self.data.contains(&(x, y)) {
            return 1;
        } else {
            return 0;
        }
    }

    fn step(&mut self) {
        let mut new_data = HashSet::new();
        for x in 0..self.width {
            for y in 0..self.height {
                let n = self.count_neighbours(x, y);
                if self.is_alive(x, y) == 1 {
                    if n < 2 || n > 3 {
                        // die
                        new_data.remove(&(x, y));
                    } else {
                        // stay alive
                        new_data.insert((x, y));
                    }
                } else if n == 3 {
                    // become alive
                    new_data.insert((x, y));
                }
            }
        }
        self.data = new_data;
    }

    /*
    fn print(&self) {
        print!("{}", "+");
        for _ in 0..self.width {
            print!("{}", "-");
        }
        println!("{}", "+");
        for y in 0..self.height {
            print!("{}", "|");
            for x in 0..self.width {
                if self.data.contains(&(x, y)) {
                    print!("{}", "O");
                } else {
                    print!("{}", " ");
                }
            }
            println!("{}", "|");
        }
        print!("{}", "+");
        for _ in 0..self.width {
            print!("{}", "-");
        }
        println!("{}", "+");
    }
    */
}

fn main() {
    let opt = Opt::from_args();
    let mut field = Field::with_size(opt.width, opt.height); //, opt.pixel_size);
    if field.read_file().is_ok() {
        field.step();
    } else {
        field.fill_random();
    }

    field.write_file();
    field.write_image(&"/tmp/gol.png");
}
